# coding=utf-8
"""
Contains the interface for querying data from the data fetcher module.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from dataclasses import dataclass, field
from typing import List, Any, Dict, Optional

from fastlane_bot.config import Config
from fastlane_bot.helpers.poolandtokens import PoolAndTokens


@dataclass
class Token:

    __VERSION__ = "0.0.1"
    __DATE__ = "2023-07-03"

    symbol: str
    address: str
    decimals: int
    key: str

    def __eq__(self, other):
        return self.key == other.key if isinstance(other, Token) else False

    def __hash__(self):
        return hash(self.key)


@dataclass
class Pool(PoolAndTokens):

    __VERSION__ = "0.0.1"
    __DATE__ = "2023-07-03"

    pass


@dataclass
class QueryInterface:
    """
    Interface for querying data from the data fetcher module. These methods mirror the existing methods
    expected/used in the bot module. The implementation of these methods should allow for the bot module
    to be used with the new data fetcher module without any changes to the bot module.
    """

    __VERSION__ = "0.0.1"
    __DATE__ = "2023-07-03"

    mgr: Any = None
    state: List[Dict[str, Any]] = field(default_factory=list)
    ConfigObj: Config = None
    uniswap_v2_event_mappings: Dict[str, str] = field(default_factory=dict)
    exchanges: List[str] = field(default_factory=list)

    @property
    def cfg(self) -> Config:
        return self.ConfigObj

    def remove_unsupported_exchanges(self) -> None:
        initial_state = self.state.copy()
        self.state = [
            pool for pool in self.state if pool["exchange_name"] in self.exchanges
        ]
        self.cfg.logger.info(
            f"Removed {len(initial_state) - len(self.state)} unsupported exchanges. {len(self.state)} pools remaining"
        )

        # Log the total number of pools remaining for each exchange
        self.ConfigObj.logger.info("Pools remaining per exchange:")
        for exchange_name in self.exchanges:
            pools = self.filter_pools(exchange_name)
            self.log_pool_numbers(pools, exchange_name)

    def has_balance(self, pool: Dict[str, Any], key: str) -> bool:
        """
        Check if a pool has a balance for a given key

        Parameters
        ----------
        pool: Dict[str, Any]
            The pool to check
        key: str
            The key to check for a balance

        Returns
        -------
        bool
            True if the pool has a balance for the given key, False otherwise

        """
        return key in pool and pool[key] > 0

    def filter_pools(self, exchange_name: str, key: str = "") -> List[Dict[str, Any]]:
        """
        Filter pools by exchange name and key

        Parameters
        ----------
        exchange_name: str
            The exchange name to filter by
        key: str
            The key to filter by

        Returns
        -------
        List[Dict[str, Any]]
            The filtered pools
        """
        if key:
            return [
                pool
                for pool in self.state
                if pool["exchange_name"] == exchange_name
                and self.has_balance(pool, key)
            ]
        else:
            return [
                pool for pool in self.state if pool["exchange_name"] == exchange_name
            ]

    def log_pool_numbers(self, pools: List[Dict[str, Any]], exchange_name: str) -> None:
        """
        Log the number of pools for a given exchange name

        Parameters
        ----------
        pools: List[Dict[str, Any]]
            The pools to log
        exchange_name: str
            The exchange name to log

        """
        self.cfg.logger.info(f"{exchange_name}: {len(pools)}")

    def remove_zero_liquidity_pools(self) -> None:
        """
        Remove pools with zero liquidity.
        """
        initial_state = self.state.copy()

        exchanges = [
            "uniswap_v3",
            "sushiswap_v2",
            "uniswap_v2",
            "bancor_v2",
            "bancor_v3",
            "carbon_v1",
        ]
        keys = [
            "liquidity",
            "tkn0_balance",
            "tkn0_balance",
            "tkn0_balance",
            "tkn0_balance",
            "",
        ]

        self.state = [
            pool
            for exchange, key in zip(exchanges, keys)
            for pool in self.filter_pools(exchange, key)
        ]

        for exchange in exchanges:
            self.log_pool_numbers(
                [pool for pool in self.state if pool["exchange_name"] == exchange],
                exchange,
            )

        zero_liquidity_pools = [
            pool for pool in initial_state if pool not in self.state
        ]

        for exchange in exchanges:
            self.log_pool_numbers(
                [
                    pool
                    for pool in zero_liquidity_pools
                    if pool["exchange_name"] == exchange
                ],
                f"{exchange}_zero_liquidity_pools",
            )

    def remove_unmapped_uniswap_v2_pools(self) -> None:
        """
        Remove unmapped uniswap_v2 pools
        """
        initial_state = self.state.copy()
        self.state = [
            pool
            for pool in self.state
            if pool["exchange_name"] != "uniswap_v2"
            or (
                pool["exchange_name"] in ["uniswap_v2", "sushiswap_v2"]
                and pool["address"] in self.uniswap_v2_event_mappings
            )
        ]
        self.cfg.logger.info(
            f"Removed {len(initial_state) - len(self.state)} unmapped uniswap_v2/sushi pools. {len(self.state)} uniswap_v2/sushi pools remaining"
        )
        self.log_umapped_pools_by_exchange(initial_state)

    def log_umapped_pools_by_exchange(self, initial_state):
        # Log the total number of pools filtered out for each exchange
        self.ConfigObj.logger.info("Unmapped uniswap_v2/sushi pools:")
        unmapped_pools = [pool for pool in initial_state if pool not in self.state]
        assert len(unmapped_pools) == len(initial_state) - len(self.state)
        uniswap_v3_unmapped = [
            pool for pool in unmapped_pools if pool["exchange_name"] == "uniswap_v3"
        ]
        self.log_pool_numbers(uniswap_v3_unmapped, "uniswap_v3")
        uniswap_v2_unmapped = [
            pool for pool in unmapped_pools if pool["exchange_name"] == "uniswap_v2"
        ]
        self.log_pool_numbers(uniswap_v2_unmapped, "uniswap_v2")
        sushiswap_v2_unmapped = [
            pool for pool in unmapped_pools if pool["exchange_name"] == "sushiswap_v2"
        ]
        self.log_pool_numbers(sushiswap_v2_unmapped, "sushiswap_v2")

    def remove_faulty_token_pools(self) -> None:
        """
        Remove pools with faulty tokens
        """
        self.cfg.logger.info(
            f"Total number of pools. {len(self.state)} before removing faulty token pools"
        )

        safe_pools = []
        for pool in self.state:
            try:
                self.get_token(pool["tkn0_key"])
                self.get_token(pool["tkn1_key"])
                safe_pools.append(pool)
            except Exception as e:
                self.cfg.logger.info(f"Exception: {e}")
                self.cfg.logger.info(
                    f"Removing pool for exchange={pool['pair_name']}, pair_name={pool['pair_name']} token={pool['tkn0_key']} from state for faulty token"
                )

        self.state = safe_pools

    @staticmethod
    def cleanup_token_key(token_key: str) -> str:
        """
        Cleanup token key. This renames keys that have more than 1 '-' in them.

        Parameters
        ----------
        token_key: str
            The token key to cleanup

        Returns
        -------
        str
            The cleaned up token key

        """
        split_key = token_key.split("-", 2)
        return (
            f"{split_key[0]}_{split_key[1]}-{split_key[2]}"
            if len(split_key) > 2
            else token_key
        )

    def handle_token_key_cleanup(self) -> None:
        """
        Cleanup token keys in state
        """
        for idx, pool in enumerate(self.state):
            key0 = self.cleanup_token_key(pool["tkn0_key"])
            key1 = self.cleanup_token_key(pool["tkn1_key"])
            self.state[idx]["tkn0_key"] = key0
            self.state[idx]["tkn1_key"] = key1
            self.state[idx]["pair_name"] = key0 + "/" + key1

    def update_state(self, state: List[Dict[str, Any]]) -> None:
        """
        Update the state.

        Parameters
        ----------
        state: List[Dict[str, Any]]
            The new state

        """
        self.state = state.copy()
        if self.state == state:
            self.cfg.logger.warning("WARNING: State not updated")

    def drop_all_tables(self) -> None:
        """
        Drop all tables. Deprecated.
        """
        raise DeprecationWarning("Method not implemented")

    def get_pool_data_with_tokens(self) -> List[PoolAndTokens]:
        """
        Get pool data with tokens
        """
        return [
            self.create_pool_and_tokens(idx, record)
            for idx, record in enumerate(self.state)
        ]

    def create_pool_and_tokens(self, idx: int, record: Dict[str, Any]) -> PoolAndTokens:
        """
        Create a pool and tokens object from a record

        Parameters
        ----------
        idx: int
            The index of the record
        record: Dict[str, Any]
            The record

        Returns
        -------
        PoolAndTokens
            The pool and tokens object

        """
        result = PoolAndTokens(
            ConfigObj=self.ConfigObj,
            id=idx,
            **{
                key: record.get(key)
                for key in [
                    "cid",
                    "last_updated",
                    "last_updated_block",
                    "descr",
                    "pair_name",
                    "exchange_name",
                    "fee",
                    "fee_float",
                    "tkn0_balance",
                    "tkn1_balance",
                    "z_0",
                    "y_0",
                    "A_0",
                    "B_0",
                    "z_1",
                    "y_1",
                    "A_1",
                    "B_1",
                    "sqrt_price_q96",
                    "tick",
                    "tick_spacing",
                    "liquidity",
                    "address",
                    "anchor",
                    "tkn0",
                    "tkn1",
                    "tkn0_key",
                    "tkn1_key",
                    "tkn0_address",
                    "tkn0_decimals",
                    "tkn1_address",
                    "tkn1_decimals",
                ]
            },
        )
        result.tkn0 = result.pair_name.split("/")[0].split("-")[0]
        result.tkn1 = result.pair_name.split("/")[1].split("-")[0]
        result.tkn0_key = result.pair_name.split("/")[0]
        result.tkn1_key = result.pair_name.split("/")[1]
        return result

    def get_tokens(self) -> List[Token]:
        """
        Get tokens. This method returns a list of tokens that are in the state.

        Returns
        -------
        List[Token]
            The list of tokens
        """
        token_set = set()
        for record in self.state:
            token_set.add(self.create_token(record, "tkn0_"))
            token_set.add(self.create_token(record, "tkn1_"))
        return list(token_set)

    def create_token(self, record: Dict[str, Any], prefix: str) -> Token:
        """
        Create a token from a record

        Parameters
        ----------
        record: Dict[str, Any]
            The record
        prefix: str
            The prefix of the token

        Returns
        -------
        Token
            The token

        """
        return Token(
            symbol=record.get(f"{prefix}symbol"),
            decimals=record.get(f"{prefix}decimals"),
            key=record.get(f"{prefix}key"),
            address=record.get(f"{prefix}address"),
        )

    def get_bnt_price_from_tokens(self, price: float, tkn: Token) -> float:
        """
        Get the BNT price from tokens

        Parameters
        ----------
        price: float
            The price
        tkn: Token
            The token

        Returns
        -------
        float
            The BNT price

        """
        raise DeprecationWarning("Method not implemented")

    def get_token(self, key: str) -> Optional[Token]:
        """
        Get a token from the state

        Parameters
        ----------
        key: str
            The token key

        Returns
        -------
        Optional[Token]
            The token

        """
        tokens = self.get_tokens()
        if "-" in key:
            return next((tkn for tkn in tokens if tkn.key == key), None)
        elif key.startswith("0x"):
            return next((tkn for tkn in tokens if tkn.address == key), None)
        else:
            raise ValueError(f"[get_token] Invalid token: {key}")

    def get_pool(self, **kwargs) -> Optional[PoolAndTokens]:
        """
        Get a pool from the state

        Parameters
        ----------
        kwargs: Dict[str, Any]
            The pool parameters

        Returns
        -------
        Pool
            The pool

        """
        pool_data_with_tokens = self.get_pool_data_with_tokens()
        try:
            pool = next(
                (
                    pool
                    for pool in pool_data_with_tokens
                    if all(getattr(pool, key) == kwargs[key] for key in kwargs)
                ),
                None,
            )
            pool.exchange_name
        except AttributeError:
            if 'cid' in kwargs:
                kwargs['cid'] = int(kwargs['cid'])
                pool = next(
                    (
                        pool
                        for pool in pool_data_with_tokens
                        if all(getattr(pool, key) == kwargs[key] for key in kwargs)
                    ),
                    None,
                )
        return pool

    def get_pools(self) -> List[PoolAndTokens]:
        """
        Get all pools from the state

        Returns
        -------
        List[Pool]
            The list of pools

        """
        return self.get_pool_data_with_tokens()

    def update_recently_traded_pools(self, cids: List[str]):
        """
        Update recently traded pools. Deprecated.

        Parameters
        ----------
        cids: List[str]
            The list of cids

        """
        raise DeprecationWarning("Method not implemented")



"""
simple representation of a pair of tokens, used by cpc and arbgraph

(c) Copyright Bprotocol foundation 2023. 
Licensed under MIT
"""
__VERSION__ = "2.0"
__DATE__ = "5/May/2023"

from dataclasses import dataclass, field, asdict, InitVar


@dataclass
class SimplePair:
    """
    a pair in notation TKNB/TKNQ; can also be provided as list (but NOT: tknb=, tknq=)
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    tknb: str = field(init=False)
    tknq: str = field(init=False)
    pair: InitVar[str] = None

    def __post_init__(self, pair):
        if isinstance(pair, self.__class__):
            self.tknb = pair.tknb
            self.tknq = pair.tknq
        elif isinstance(pair, str):
            self.tknb, self.tknq = pair.split("/")
        elif pair is False:
            # used in alternative constructors
            pass
        else:
            try:
                self.tknb, self.tknq = pair
            except:
                raise ValueError(f"pair must be a string or list of two strings {pair}")

    @classmethod
    def from_tokens(cls, tknb, tknq):
        pair = cls(False)
        pair.tknb = tknb
        pair.tknq = tknq
        return pair

    def __str__(self):
        return f"{self.tknb}/{self.tknq}"

    @property
    def pair(self):
        """string representation of the pair"""
        return str(self)

    @property
    def pairt(self):
        """tuple representation of the pair"""
        return (self.tknb, self.tknq)

    @property
    def pairr(self):
        """returns the reversed pair"""
        return f"{self.tknq}/{self.tknb}"

    @property
    def pairrt(self):
        """tuple representation of the reverse pair"""
        return (self.tknq, self.tknb)

    @property
    def tknx(self):
        return self.tknb

    @property
    def tkny(self):
        return self.tknq

    NUMERAIRE_TOKENS = {
        tkn: i
        for i, tkn in enumerate(
            [
                "USDC",
                "USDT",
                "DAI",
                "TUSD",
                "BUSD",
                "PAX",
                "GUSD",
                "USDS",
                "sUSD",
                "mUSD",
                "HUSD",
                "USDN",
                "USDP",
                "USDQ",
                "ETH",
                "WETH",
                "WBTC",
                "BTC",
            ]
        )
    }

    @classmethod
    def n(cls, tkn):
        """normalize the token name (remove the id, if any)"""
        if len(tkn.split("/")) > 1:
            return "/".join([cls.n(t) for t in tkn.split("/")])
        return tkn.split("-")[0].split("(")[0]

    @property
    def tknb_n(self):
        return self.n(self.tknb)

    @property
    def tknq_n(self):
        return self.n(self.tknq)
    
    @property
    def pair_n(self):
        """normalized pair"""
        return f"{self.tknb_n}/{self.tknq_n}"

    @property
    def tknx_n(self):
        return self.n(self.tknx)

    @property
    def tkny_n(self):
        return self.n(self.tkny)

    @property
    def isprimary(self):
        """whether the representation is primary or secondary"""
        tknqix = self.NUMERAIRE_TOKENS.get(self.tknq_n, 1e10)
        tknbix = self.NUMERAIRE_TOKENS.get(self.tknb_n, 1e10)
        if tknqix == tknbix:
            return self.tknb < self.tknq
        return tknqix < tknbix

    def primary_price(self, p):
        """returns the primary price (p if primary, 1/p if secondary)"""
        if self.isprimary:
            return p
        else:
            if p == 0:
                return float("nan")
        return 1 / p
    pp = primary_price
    
    @property
    def pp_convention(self):
        """returns the primary price convention"""
        tknb, tknq = self.primary_n.split("/")
        return f"{tknq} per {tknb}"

    @property
    def primary(self):
        """returns the primary pair"""
        return self.pair if self.isprimary else self.pairr
    
    @property
    def primary_n(self):
        """the primary pair, normalized"""
        tokens = self.primary.split("/")
        tokens = [self.n(t) for t in tokens]
        return "/".join(tokens)
    
    @property
    def primary_tknb(self):
        """returns the primary normailised tknb"""
        return self.tknb_n if self.isprimary else self.tknq_n

    @property
    def primary_tknq(self):
        """returns the primary normailised tknq"""
        return self.tknq_n if self.isprimary else self.tknb_n
    
    @property
    def secondary(self):
        """returns the secondary pair"""
        return self.pairr if self.isprimary else self.pair
    
    @property
    def secondary_n(self):
        """the secondary pair, normalized"""
        tokens = self.secondary.split("/")
        tokens = [self.n(t) for t in tokens]
        return "/".join(tokens)

    @classmethod
    def wrap(cls, pairlist):
        """wraps a list of strings into Pairs"""
        return tuple(cls(p) for p in pairlist)

    @classmethod
    def unwrap(cls, pairlist):
        """unwraps a list of Pairs into strings"""
        return tuple(str(p) for p in pairlist)



"""
Fastlane bot config -- provider
"""
__VERSION__ = "0.9.1"
__DATE__ = "30/Apr 2023"
from .base import ConfigBase
from . import selectors as S
from .network import ConfigNetwork
from .connect import EthereumNetwork
import os
from dotenv import load_dotenv
load_dotenv()

from ..data.abi import (
    BANCOR_V3_NETWORK_INFO_ABI,
    CARBON_CONTROLLER_ABI,
    FAST_LANE_CONTRACT_ABI,
)
from brownie import Contract

ETH_PRIVATE_KEY_BE_CAREFUL = os.environ.get("ETH_PRIVATE_KEY_BE_CAREFUL")
WEB3_ALCHEMY_PROJECT_ID = os.environ.get("WEB3_ALCHEMY_PROJECT_ID")

class ConfigProvider(ConfigBase):
    """
    Fastlane bot config -- provider
    """
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    
    RPC_URL = None # set in derived class init
    
    PROVIDER_DEFAULT = S.PROVIDER_DEFAULT
    PROVIDER_INFURA = S.PROVIDER_INFURA
    PROVIDER_ALCHEMY = S.PROVIDER_ALCHEMY
    PROVIDER_TENDERLY = S.PROVIDER_TENDERLY
    PROVIDER_UNITTEST = S.PROVIDER_UNITTEST
    ETH_PRIVATE_KEY_BE_CAREFUL = ETH_PRIVATE_KEY_BE_CAREFUL
    WEB3_ALCHEMY_PROJECT_ID = WEB3_ALCHEMY_PROJECT_ID


    @classmethod
    def new(cls, network: ConfigNetwork, provider=None, **kwargs):
        """
        Return a new ConfigProvider.
        """
        if not issubclass(type(network), ConfigNetwork):
            raise TypeError(f"Invalid network type: {type(network)}")
        
        if provider is None:
            provider = cls.PROVIDER_DEFAULT
            
        if provider == S.PROVIDER_DEFAULT:
            provider = network.DEFAULT_PROVIDER
        
        if provider == S.PROVIDER_ALCHEMY:
            return _ConfigProviderAlchemy(network, _direct=False, **kwargs)
        elif provider == S.PROVIDER_TENDERLY:
            return _ConfigProviderTenderly(network, _direct=False, **kwargs)
        elif provider == S.PROVIDER_INFURA:
            return _ConfigProviderInfura(network, _direct=False, **kwargs)
        elif provider == S.PROVIDER_UNITTEST:
            return _ConfigProviderUnitTest(network, _direct=False, **kwargs)
        else:
            raise ValueError(f"Unknown provider: {provider}")
        
    def __init__(self, network: ConfigNetwork, **kwargs):
        super().__init__(**kwargs)
        self.network = network


class _ConfigProviderAlchemy(ConfigProvider):
    """
    Fastlane bot config -- provider [Alchemy]
    """
    PROVIDER = S.PROVIDER_ALCHEMY
    WEB3_ALCHEMY_PROJECT_ID = WEB3_ALCHEMY_PROJECT_ID
    
    def __init__(self, network: ConfigNetwork, **kwargs):
        super().__init__(network, **kwargs)
        assert self.network.NETWORK == ConfigNetwork.NETWORK_ETHEREUM, f"Alchemy only supports Ethereum {self.network}"
        self.RPC_URL = f"https://eth-mainnet.alchemyapi.io/v2/{self.WEB3_ALCHEMY_PROJECT_ID}"

        N = self.network
        self.connection = EthereumNetwork(
            network_id=N.NETWORK_ID,
            network_name=f"{N.NETWORK_NAME} (Alchemy)",
            provider_url=self.RPC_URL,
            provider_name="alchemy",
        )
        self.connection.connect_network()
        self.w3 = self.connection.web3
        self.LOCAL_ACCOUNT = self.w3.eth.account.from_key(ETH_PRIVATE_KEY_BE_CAREFUL)
        self.BANCOR_NETWORK_INFO_CONTRACT = Contract.from_abi(
            name=N.BANCOR_V3_NAME,
            address=N.BANCOR_V3_NETWORK_INFO_ADDRESS,
            abi=BANCOR_V3_NETWORK_INFO_ABI,
        )
        self.CARBON_CONTROLLER_CONTRACT = Contract.from_abi(
            name=N.CARBON_V1_NAME,
            address=N.CARBON_CONTROLLER_ADDRESS,
            abi=CARBON_CONTROLLER_ABI,
        )
        self.BANCOR_ARBITRAGE_CONTRACT = self.w3.eth.contract(
            address=self.w3.toChecksumAddress(N.FASTLANE_CONTRACT_ADDRESS),
            abi=FAST_LANE_CONTRACT_ABI,
        )
        reward_percent, max_profit = self.BANCOR_ARBITRAGE_CONTRACT.caller.rewards()

        self.ARB_REWARD_PERCENTAGE = str(int(reward_percent) / 1000000)
        self.ARB_MAX_PROFIT = str(int(max_profit) / (10 ** 18))
        self.EXPECTED_GAS_MODIFIER = "0.85"


class _ConfigProviderTenderly(ConfigProvider):
    """
    Fastlane bot config -- provider [Tenderly]
    """
    PROVIDER = S.PROVIDER_TENDERLY
    
    def __init__(self, network: ConfigNetwork, **kwargs):
        super().__init__(network, **kwargs)
        assert self.network.NETWORK == ConfigNetwork.NETWORK_TENDERLY, f"Tenderly only supports Tenderly {self.network}"
        self.RPC_URL = f"https://rpc.tenderly.co/fork/{self.network.TENDERLY_FORK}"

        N = self.network
        self.connection = EthereumNetwork(
            network_id=N.NETWORK_ID,
            network_name=f"{N.NETWORK_NAME}",
            provider_url=self.RPC_URL,
            provider_name="tenderly",
        )
        self.connection.connect_network()
        self.w3 = self.connection.web3
        self.LOCAL_ACCOUNT = self.w3.eth.account.from_key(ETH_PRIVATE_KEY_BE_CAREFUL)

        self.BANCOR_NETWORK_INFO_CONTRACT = Contract.from_abi(
            name=N.BANCOR_V3_NAME,
            address=N.BANCOR_V3_NETWORK_INFO_ADDRESS,
            abi=BANCOR_V3_NETWORK_INFO_ABI,
        )
        self.CARBON_CONTROLLER_CONTRACT = Contract.from_abi(
            name=N.CARBON_V1_NAME,
            address=N.CARBON_CONTROLLER_ADDRESS,
            abi=CARBON_CONTROLLER_ABI,
        )
        self.BANCOR_ARBITRAGE_CONTRACT = self.w3.eth.contract(
            address=self.w3.toChecksumAddress(N.FASTLANE_CONTRACT_ADDRESS),
            abi=FAST_LANE_CONTRACT_ABI,
        )
        reward_percent, max_profit = self.BANCOR_ARBITRAGE_CONTRACT.caller.rewards()

        self.ARB_REWARD_PERCENTAGE = str(int(reward_percent) / 1000000)
        self.ARB_MAX_PROFIT = str(int(max_profit) / (10 ** 18))

    
class _ConfigProviderInfura(ConfigProvider):
    """
    Fastlane bot config -- provider [Infura]
    """
    PROVIDER = S.PROVIDER_INFURA
    
    def __init__(self, network: ConfigNetwork, **kwargs):
        super().__init__(network, **kwargs)
        assert self.network.NETWORK == ConfigNetwork.NETWORK_ETHEREUM, f"Alchemy only supports Ethereum {self.network}"
        raise NotImplementedError("Infura not implemented")

class _ConfigProviderUnitTest(ConfigProvider):
    """
    Fastlane bot config -- provider [UnitTest]
    """
    PROVIDER = S.PROVIDER_UNITTEST
    
    def __init__(self, network: ConfigNetwork, **kwargs):
        super().__init__(network, **kwargs)
        #assert self.network.NETWORK == ConfigNetwork.NETWORK_ETHEREUM, f"Alchemy only supports Ethereum {self.network}"
        #raise NotImplementedError("Infura not implemented")
        self.connection = None
        self.w3 = None
        self.BANCOR_NETWORK_INFO_CONTRACT = None
        self.CARBON_CONTROLLER_CONTRACT = None
        self.BANCOR_ARBITRAGE_CONTRACT = None
        

    



"""
Carbon helper module - parameter management
"""
__VERSION__ = "1.2"
__DATE__ = "29/01/2023"


class Params:
    """
    parameter management

    EXAMPLE

        # Standard

        p = Params(a=1, b=2)
        p.a            # 1
        p["b"]         # 2
        p.c            # raises; must exists when accessed as attribute
        p["c"]         # None; fails gracefully when accessed via []
        p["c"] = 3     # OK
        p.c            # 3; after assignment
        p["c"]         # 3; after assignment
        p["b"] = 3     # raises; re-assignment not allowed

        p = Params(**{"a": 1, "b": 2})  # creating params from dict

        # With defaults

        p = Params(a=1, b=2)
        p.set_default(**{"b":20, "c":3})
        p.a                           # 1; from params
        p.b                           # 2; from params
        p.c                           # 3; from defaults


    """

    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    def __init__(self, **kwargs):
        self._params = dict(kwargs)
        self._defaults = None

    @classmethod
    def construct(cls, dct=None, defaults=None):
        """
        alternative constructor from dct

        :dct:       typically a dict object; can also be a Params object which will be replicated
                    (defaults can either be on the original object or here, but not on both; if
                    you want to merge defaults use set_default on the created object); a value
                    of None creates an empty object
        :defaults:  the default values for this object; note that they can not be passed
                    using the standard constructor; if the object is already a params object,
                    the existing defaults will be updated, and overwritten if they exist
        :returns:   a newly created Params object
        """
        if not dct:
            result = cls()
        elif isinstance(dct, cls):
            result = cls(**dct._params)
            if not dct._defaults is None and not defaults is None:
                raise ValueError(
                    "Must not provide default in both constructor and dct",
                    dct,
                    defaults,
                )
        else:
            result = cls(**dct)

        if defaults:
            result._defaults = {**defaults}
        return result

    def add(self, **kwargs):
        """
        adds additional parameters from kwargs (params must not yet exist)

        :returns:     self (for chaining)
        """
        for k, v in kwargs.items():
            self[k] = v
        return self

    def get_default(self, item, raiseonerror=False):
        """
        gets the default value (None if does not exist)
        """
        if self._defaults is None:
            self.set_default()
        if raiseonerror:
            return self._defaults[item]
        else:
            return self._defaults.get(item, None)

    def set_default(self, **kwargs):
        """
        adds default params

        :returns:    self for chaining
        """
        if self._defaults is None:
            self._defaults = dict()
        else:
            for k, v in kwargs.items():
                self._defaults[k] = v
        return self

    @property
    def params(self):
        """
        returns the parameters as dict
        """
        return self._params

    @property
    def defaults(self):
        """
        returns defaults object (creates empty one if it does not exist)
        """
        if self._defaults is None:
            self.set_default()
        return self._defaults

    def set(self, item, value, allowupdate=True):
        """
        sets an item

        :item:          the item to be set
        :value:         the value to set it to
        :allowupdate:   if True (default), existing items can be changed
        :returns:       self (for chaining)
        """
        if not allowupdate:
            if item in self._params:
                raise ValueError(
                    f"Item {item} already exists with value {self._params[item]} and update not allowed.",
                    value,
                    self._params,
                    allowupdate,
                )
        self._params[item] = value
        return self

    def __getitem__(self, item):
        try:
            return self._params[item]
        except KeyError:
            return self.get_default(item, raiseonerror=False)

    def __setitem__(self, item, value):
        self.set(item, value, allowupdate=False)

    def __getattr__(self, item):
        """
        for all item starting with _ this refers to super().__getattr__
        """
        if item[:1] == "_":
            return super().__getattr__(item)
        try:
            return self._params[item]
        except KeyError:
            return self.get_default(item, raiseonerror=True)

    def __repr__(self):

        defaults = f", defaults={self._defaults}" if self._defaults else ""
        return f"{self.__class__.__name__}.construct({self._params}{defaults})"



# coding=utf-8
"""
Base class for triangular arbitrage finder modes

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import abc
import itertools
from typing import List, Any, Tuple, Union

import pandas as pd

from fastlane_bot.modes.base import ArbitrageFinderBase
from fastlane_bot.tools.cpc import T


class ArbitrageFinderTriangleBase(ArbitrageFinderBase):
    """
    Base class for triangular arbitrage finder modes
    """

    @abc.abstractmethod
    def find_arbitrage(self, candidates: List[Any] = None, ops: Tuple = None, best_profit: float = 0, profit_src: float = 0) -> Union[List, Tuple]:
        """
        see base.py
        """
        pass

    @staticmethod
    def get_miniverse(
        y_match_curves_not_carbon: List[Any],
        base_exchange_curves: List[Any],
        x_match_curves_not_carbon: List[Any],
        flt: str,
        arb_mode: str,
        combos: List[Any],
    ) -> List[Any]:
        """
        Get miniverse for triangular arbitrage

        Parameters
        ----------
        y_match_curves_not_carbon : list
            List of curves that match the y token and are not on carbon
        base_exchange_curves : list
            List of curves on the base exchange
        x_match_curves_not_carbon : list
            List of curves that match the x token and are not on carbon
        flt : str
            Flashloan token
        arb_mode : str
            Arbitrage mode
        combos : list
            List of combos

        Returns
        -------
        combos : list
            List of combos

        """
        if arb_mode in ["single_triangle", "triangle"]:
            miniverses = list(
                itertools.product(
                    y_match_curves_not_carbon,
                    base_exchange_curves,
                    x_match_curves_not_carbon,
                )
            )
        else:
            external_curve_combos = list(
                itertools.product(y_match_curves_not_carbon, x_match_curves_not_carbon)
            )
            miniverses = [
                base_exchange_curves + list(combo) for combo in external_curve_combos
            ]
        if miniverses:
            combos += list(zip([flt] * len(miniverses), miniverses))
        return combos

    def get_combos(
        self, flashloan_tokens: List[str], CCm: Any, arb_mode: str
    ) -> Tuple[List[str], List[Any]]:
        """
        Get combos for triangular arbitrage

        Parameters
        ----------
        flashloan_tokens : list
            List of flashloan tokens
        CCm : object
            CCm object
        arb_mode : str
            Arbitrage mode

        Returns
        -------
        combos : list
            List of combos

        """
        combos = []
        if arb_mode in ["single_triangle_bancor3", "bancor_v3", "b3_two_hop"]:
            combos = [
                (tkn0, tkn1)
                for tkn0, tkn1 in itertools.product(flashloan_tokens, flashloan_tokens)
                # note that the pair is tkn0/tkn1, ie tkn1 is the quote token
                if tkn0 != tkn1
            ]
        else:
            all_base_exchange_curves = CCm.byparams(exchange=self.base_exchange).curves
            for flt in flashloan_tokens:  # may wish to run this for one flt at a time
                non_flt_base_exchange_curves = [
                    x for x in all_base_exchange_curves if flt not in x.pair
                ]
                for non_flt_base_exchange_curve in non_flt_base_exchange_curves:
                    target_tkny = non_flt_base_exchange_curve.tkny
                    target_tknx = non_flt_base_exchange_curve.tknx
                    base_exchange_curves = (
                        CCm.bypairs(f"{target_tknx}/{target_tkny}")
                        .byparams(exchange=self.base_exchange)
                        .curves
                    )
                    y_match_curves = CCm.bypairs(
                        set(CCm.filter_pairs(onein=target_tknx))
                        & set(CCm.filter_pairs(onein=flt))
                    )
                    x_match_curves = CCm.bypairs(
                        set(CCm.filter_pairs(onein=target_tkny))
                        & set(CCm.filter_pairs(onein=flt))
                    )
                    y_match_curves_not_carbon = [
                        x
                        for x in y_match_curves
                        if x.params.exchange != self.base_exchange
                    ]
                    x_match_curves_not_carbon = [
                        x
                        for x in x_match_curves
                        if x.params.exchange != self.base_exchange
                    ]
                    combos = self.get_miniverse(
                        y_match_curves_not_carbon,
                        base_exchange_curves,
                        x_match_curves_not_carbon,
                        flt,
                        arb_mode,
                        combos,
                    )
        return combos

    @staticmethod
    def get_mono_direction_carbon_curves(
        miniverse: List[Any], trade_instructions_df: pd.DataFrame, token_in: str=None
    ) -> List[Any]:
        """
        Get mono direction carbon curves for triangular arbitrage

        Parameters
        ----------
        miniverse : list
            List of miniverses
        token_in : str
            Token in
        trade_instructions_df : DataFrame
            Trade instructions dataframe

        Returns
        -------
        mono_direction_carbon_curves : list
            List of mono direction carbon curves

        """

        if token_in is None:
            columns = trade_instructions_df.columns
            check_nan = trade_instructions_df.copy().fillna(0)
            first_bancor_v3_pool = check_nan.iloc[0]
            second_bancor_v3_pool = check_nan.iloc[1]

            for idx, token in enumerate(columns):
                if token == T.BNT:
                    continue
                if first_bancor_v3_pool[token] < 0:
                    token_in = token
                    break
                if second_bancor_v3_pool[token] < 0:
                    token_in = token
                    break

        wrong_direction_cids = []
        for idx, row in trade_instructions_df.iterrows():
            if (row[token_in] < 0) and ("-0" in idx or "-1" in idx):
                wrong_direction_cids.append(idx)

        return [curve for curve in miniverse if curve.cid not in wrong_direction_cids]
